import importlib
import os
import re

from django.conf import settings
from django.db import connections
from django.test.utils import setup_databases


def get_db_backend(connection):
    vendor_map = {
        "sqlite": "django_migrations_ci.backends.sqlite3",
        "postgresql": "django_migrations_ci.backends.postgresql",
    }
    return importlib.import_module(vendor_map[connection.vendor])


def create_test_db():
    for connection in connections.all():
        connection.creation._create_test_db(
            verbosity=True, autoclobber=True, keepdb=False
        )


def setup_test_db():
    # Based on https://github.com/django/django/blob/d62563cbb194c420f242bfced52b37d6638e67c6/django/test/runner.py#L1051-L1054  # noqa: E501
    aliases = []
    database_names = {}
    for connection in connections.all():
        aliases.append(connection.alias)
        database_names[connection.alias] = connection.settings_dict["NAME"]

    setup_databases(verbosity=True, interactive=False, aliases=aliases)

    # Django setup_databases change original settings and don't care about it
    # because it run the setup only one time and other parts of testing understand that.
    for connection in connections.all():
        database_name = database_names[connection.alias]
        connection.close()
        settings.DATABASES[connection.alias]["NAME"] = database_name
        connection.settings_dict["NAME"] = database_name


def clone_test_db(connection, parallel, is_pytest=False):
    for index in range(parallel):
        if is_pytest:
            # pytest-django use test_db_gwN, from 0 to N-1.
            # e.g. test_db_gw0, test_db_gw1, ...
            # https://github.com/pytest-dev/pytest-django/blob/e0c77b391ea54c3b8d6ffbb593aa25188a0ce7e9/pytest_django/fixtures.py#L61  # noqa: E501
            suffix = f"gw{index}"
        else:
            # Django use test_db_N, from 1 to N.
            # e.g. test_db_1, test_db_2, ...
            suffix = f"{index + 1}"

        connection.creation.clone_test_db(suffix=suffix, verbosity=True, keepdb=False)

        if is_pytest and connection.vendor == "sqlite":
            settings_dict = connection.creation.get_test_db_clone_settings(suffix)
            django_db_name = settings_dict["NAME"]

            if "." in django_db_name:
                # Django clone_test_db create file db_gw0.sqlite3, but pytest-django
                # expects db.sqlite3_gw0. Lets rename the file.
                pytest_db_name = re.sub(r"(_gw\d+)\.(.+)$", r".\2\1", django_db_name)

                # Move db_gw0.sqlite3 to db.sqlite3_gw0.
                os.rename(django_db_name, pytest_db_name)



# Generated by Django 4.1.2 on 2022-10-20 18:43

from django.db import migrations


def create_one_bus(apps, schema_editor):
    Bus = apps.get_model("testapp", "Bus")
    Bus.objects.create()


class Migration(migrations.Migration):

    dependencies = [
        ("testapp", "0001_initial"),
    ]

    operations = [migrations.RunPython(create_one_bus, migrations.RunPython.noop)]



from contextlib import contextmanager
from pathlib import Path

from django.conf import settings
from django.core.management.base import BaseCommand
from django.db import connections
from django.test.runner import get_max_test_processes

from django_migrations_ci import django


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument("--database", default="default")
        parser.add_argument("-n", "--parallel", default=None)
        parser.add_argument(
            "--pytest",
            dest="is_pytest",
            action="store_true",
            default=False,
        )
        parser.add_argument("--directory", default="")

    def handle(self, *args, database, parallel, is_pytest, directory, **options):
        if parallel == "auto":
            parallel = get_max_test_processes()
        elif parallel is not None:
            parallel = int(parallel)

        connection = connections[database]
        cached_file = Path(directory) / Path(f"migrateci-{database}")

        backend = django.get_db_backend(connection)

        if Path(cached_file).exists():
            print("Database cache exists.")
            django.create_test_db()
            with _override_test_db(connection):
                backend.load(connection, cached_file)
        else:
            print("Database cache does not exist.")
            django.setup_test_db()
            backend.dump(connection, cached_file)

        if parallel:
            with _override_test_db(connection):
                django.clone_test_db(
                    connection=connection,
                    parallel=parallel,
                    is_pytest=is_pytest,
                )


@contextmanager
def _override_test_db(connection):
    # Django clone_test_db trust setup_databases already changed original settings,
    # so I have to do that here.
    try:
        test_db_name = connection.settings_dict["TEST"]["NAME"]
    except KeyError:
        test_db_name = None

    if not test_db_name:
        test_db_name = connection.creation._get_test_db_name()

    db_name = connection.settings_dict["NAME"]
    connection.settings_dict["NAME"] = test_db_name
    settings.DATABASES[connection.alias]["NAME"] = test_db_name
    connection.close()

    yield

    connection.settings_dict["NAME"] = db_name
    settings.DATABASES[connection.alias]["NAME"] = db_name
    connection.close()



from pathlib import Path

import pytest


def _rm(pathname):
    for filename in Path(".").glob(pathname):
        Path(filename).unlink()


@pytest.fixture(autouse=True)
def remove_sqlite3_files():
    pathname = "dbtest*.sqlite3"
    _rm(pathname)
    yield
    _rm(pathname)



from pathlib import Path

from django.core.management import execute_from_command_line


def test_migrateci():
    execute_from_command_line(["manage.py", "migrateci"])
    assert Path("dbtest.sqlite3").exists()
    assert not Path("dbtest_1.sqlite3").exists()


def test_migrateci_parallel():
    execute_from_command_line(["manage.py", "migrateci", "--parallel", "1"])
    assert Path("dbtest.sqlite3").exists()
    assert Path("dbtest_1.sqlite3").exists()
    assert not Path("dbtest_2.sqlite3").exists()


def test_migrateci_pytest():
    execute_from_command_line(
        [
            "manage.py",
            "migrateci",
            "--parallel",
            "1",
            "--pytest",
        ]
    )
    assert Path("dbtest.sqlite3").exists()
    assert Path("dbtest.sqlite3_gw0").exists()


def test_migrateci_cached(mocker):
    # Create empty cache file.
    Path("migrateci-default").touch()
    setup_test_db_mock = mocker.patch("django_migrations_ci.django.setup_test_db")
    execute_from_command_line(["manage.py", "migrateci"])
    assert Path("dbtest.sqlite3").exists()
    setup_test_db_mock.assert_not_called()



import sqlite3


def load(connection, input_file):
    db_name = connection.settings_dict["TEST"]["NAME"]

    print(f"Connecting to sqlite3 {db_name}.")
    with open(input_file, "r") as f:
        sql = f.read()
        print(f"Load SQL to {db_name}: {sql}")

        with sqlite3.connect(db_name) as conn:
            conn.executescript(sql)


def dump(connection, output_file):
    db_name = connection.settings_dict["TEST"]["NAME"]

    print(f"Connecting to sqlite3 {db_name}.")
    with sqlite3.connect(db_name) as conn:
        sql = "".join(f"{sql}\n" for sql in conn.iterdump())
        print(f"Dump SQL to {db_name}: {sql}")

        with open(output_file, "w") as f:
            f.write(sql)



from django_migrations_ci.backends.postgresql import shell


def load(connection, input_file):
    """
    psql -h $HOST -p $PORT -U $USER -d test_foo -f migrate-postgresql
    """
    shell.load(connection.settings_dict, input_file)


def dump(connection, output_file):
    """
    pg_dump -F c -h $DB_HOST -U $POSTGRES_USER test_foo -f migrateci-postgresql
    """
    shell.dump(connection.settings_dict, output_file)



import os
from pathlib import Path
import tempfile

import pytest


@pytest.fixture(scope="session", autouse=True)
def setup_env():
    os.environ["DJANGO_SETTINGS_MODULE"] = "django_migrations_ci.tests.testapp.settings"


def _rm(pathname):
    for filename in Path(".").glob(pathname):
        Path(filename).unlink()


@pytest.fixture(autouse=True)
def remove_cached_files():
    pathname = "migrateci-*"
    _rm(pathname)
    yield
    _rm(pathname)


@pytest.fixture
def tempdir():
    directory = tempfile.TemporaryDirectory()
    with directory:
        yield directory.name



from contextlib import contextmanager

from django.db import connections


@contextmanager
def _get_cursor():
    # Encapsulate connection because it must close the connection too,
    # because I can't drop database if this connection remains open.
    connection = connections["postgresql"]
    try:
        yield connection.cursor()
    finally:
        connection.close()


def databases():
    with _get_cursor() as cursor:
        cursor.execute("select datname FROM pg_database")
        dbs = {db for db, in cursor.fetchall()}
    return dbs



from subprocess import PIPE, Popen


def _exec(command, env):
    p = Popen(
        command,
        shell=True,
        stdin=PIPE,
        stdout=PIPE,
        stderr=PIPE,
        env=env,
    )
    stdout, stderr = p.communicate()

    if stderr:
        print("EXEC ERROR", command, stdout, stderr)
    return stdout, stderr


def _ctx(db_conf):
    env = {"PGPASSWORD": db_conf["PASSWORD"]}
    try:
        database = db_conf["TEST"]["NAME"] or db_conf["NAME"]
    except KeyError:
        database = db_conf["NAME"]

    data = {
        "host": db_conf["HOST"],
        "port": db_conf["PORT"],
        "user": db_conf["USER"],
        "database": database,
    }
    return data, env


def load(db_conf, input_file):
    ctx, env = _ctx(db_conf)
    command = "psql -h {host} -p {port} -U {user} -d {database} -f {input_file}"
    _exec(command.format(input_file=input_file, **ctx), env)


def dump(db_conf, output_file):
    ctx, env = _ctx(db_conf)
    pg_dump = "pg_dump -Fp -h {host} -p {port} -U {user} -d {database} -f {output_file}"
    _exec(pg_dump.format(output_file=output_file, **ctx), env)



import os
from pathlib import Path
from shutil import copyfile

from django.core.management import execute_from_command_line
import pytest

from django_migrations_ci.tests.postgresql import utils


@pytest.fixture
def db_name():
    # Do not use django settings here, because Django change it in runtime.
    return os.getenv("POSTGRES_DB", "postgres")


def test_migrateci_postgresql(db_name):
    execute_from_command_line(["manage.py", "migrateci", "--database", "postgresql"])
    assert Path("migrateci-postgresql").exists()
    databases = utils.databases()
    assert f"test_{db_name}" in databases


def test_migrateci_postgresql_parallel(db_name):
    execute_from_command_line(
        [
            "manage.py",
            "migrateci",
            "--parallel",
            "2",
            "--database",
            "postgresql",
        ]
    )
    databases = utils.databases()
    assert f"test_{db_name}" in databases
    assert f"test_{db_name}_1" in databases
    assert f"test_{db_name}_2" in databases
    assert f"test_{db_name}_3" not in databases


def test_migrateci_pytest(db_name):
    execute_from_command_line(
        [
            "manage.py",
            "migrateci",
            "--parallel",
            "1",
            "--pytest",
            "--database",
            "postgresql",
        ]
    )
    databases = utils.databases()
    assert f"test_{db_name}_gw0" in databases


def test_migrateci_cached(mocker):
    # Create empty cache file.
    basepath = Path(__file__).parent
    copyfile(basepath / "postgres_dump.sql", "migrateci-postgresql")
    setup_test_db_mock = mocker.patch("django_migrations_ci.django.setup_test_db")

    execute_from_command_line(["manage.py", "migrateci", "--database", "postgresql"])
    setup_test_db_mock.assert_not_called()


def test_migrateci_postgresql_custom_directory(tempdir, db_name):
    execute_from_command_line(
        [
            "manage.py",
            "migrateci",
            "--database",
            "postgresql",
            "--directory",
            tempdir,
        ]
    )
    assert Path(f"{tempdir}/migrateci-postgresql").exists()
    databases = utils.databases()
    assert f"test_{db_name}" in databases
